import argparse
import os
import re
import sys
import time

import pytest
import responses

from internetarchive import get_item
from internetarchive.cli.ia_download import (
    build_range_jobs,
    normalize_byte_range,
    parse_byte_ranges,
)
from internetarchive.utils import json
from tests.conftest import (
    NASA_EXPECTED_FILES,
    IaRequestsMock,
    call_cmd,
    files_downloaded,
    ia_call,
    load_test_data_file,
)


def test_no_args(tmpdir_ch):
    call_cmd('ia --insecure download nasa')
    assert files_downloaded(path='nasa') == NASA_EXPECTED_FILES


@pytest.mark.xfail("CI" in os.environ, reason="May timeout on continuous integration")
def test_https(tmpdir_ch):
    call_cmd('ia download nasa')
    assert files_downloaded(path='nasa') == NASA_EXPECTED_FILES


def test_dry_run():
    nasa_url = 'http://archive.org/download/nasa/'
    expected_urls = {nasa_url + f for f in NASA_EXPECTED_FILES}

    stdout, _stderr = call_cmd('ia --insecure download --dry-run nasa')
    output_lines = stdout.split('\n')
    dry_run_urls = {x.strip() for x in output_lines if x and 'nasa:' not in x}

    assert expected_urls == dry_run_urls


def test_glob(tmpdir_ch):
    expected_files = {
        'globe_west_540.jpg',
        'globe_west_540_thumb.jpg',
        'nasa_itemimage.jpg',
        '__ia_thumb.jpg',
    }

    call_cmd('ia --insecure download --glob="*jpg" nasa')
    assert files_downloaded(path='nasa') == expected_files


def test_exclude(tmpdir_ch):
    expected_files = {
        'globe_west_540.jpg',
        'nasa_itemimage.jpg',
    }

    call_cmd('ia --insecure download --glob="*jpg" --exclude="*thumb*" nasa')
    assert files_downloaded(path='nasa') == expected_files


def _dry_run_filenames(capsys):
    out, _err = capsys.readouterr()
    return {line.split('/')[-1] for line in out.split('\n') if line and 'nasa:' not in line}


def test_glob_repeated(tmpdir_ch, capsys):
    """`--glob` repeated multiple times should union the patterns."""
    with IaRequestsMock() as mocker:
        mocker.add_metadata_mock('nasa')
        ia_call(['ia', '--insecure', 'download', '--dry-run',
                 '--glob', '*meta.xml', '--glob', '*reviews.xml', 'nasa'])
    assert _dry_run_filenames(capsys) == {'nasa_meta.xml', 'nasa_reviews.xml'}


def test_glob_mixed_pipe_and_repeated(tmpdir_ch, capsys):
    """`--glob "a|b" --glob c` should match the union of all three patterns."""
    with IaRequestsMock() as mocker:
        mocker.add_metadata_mock('nasa')
        ia_call(['ia', '--insecure', 'download', '--dry-run',
                 '--glob', '*meta.xml|*reviews.xml', '--glob', '*.torrent', 'nasa'])
    assert _dry_run_filenames(capsys) == {
        'nasa_meta.xml', 'nasa_reviews.xml', 'nasa_archive.torrent',
    }


def test_exclude_repeated(tmpdir_ch, capsys):
    """`--exclude` repeated multiple times should union the excludes."""
    with IaRequestsMock() as mocker:
        mocker.add_metadata_mock('nasa')
        ia_call(['ia', '--insecure', 'download', '--dry-run',
                 '--glob', '*.xml',
                 '--exclude', '*reviews*', '--exclude', '*files*',
                 'nasa'])
    assert _dry_run_filenames(capsys) == {'nasa_meta.xml'}


def test_format(tmpdir_ch):
    call_cmd('ia --insecure download --format="Archive BitTorrent" nasa')
    assert files_downloaded(path='nasa') == {'nasa_archive.torrent'}


def test_on_the_fly_format():
    i = 'wonderfulwizardo00baumiala'

    stdout, _stderr = call_cmd(f'ia --insecure download --dry-run --format="DAISY" {i}')
    assert stdout == ''

    stdout, _stderr = call_cmd(f'ia --insecure download --dry-run --format="DAISY" --on-the-fly {i}')
    assert stdout == f'http://archive.org/download/{i}/{i}_daisy.zip'


def test_clobber(tmpdir_ch):
    cmd = 'ia --insecure download nasa nasa_meta.xml'
    call_cmd(cmd)
    assert files_downloaded('nasa') == {'nasa_meta.xml'}

    _stdout, stderr = call_cmd(cmd)
    assert files_downloaded('nasa') == {'nasa_meta.xml'}
    prefix = 'nasa:\n'.replace('\n', os.linesep)
    filepath = os.path.join('nasa', 'nasa_meta.xml')
    expected_stderr = f'{prefix} skipping {filepath}, file already exists based on length and date.'
    assert expected_stderr == stderr


def test_checksum(tmpdir_ch):
    call_cmd('ia --insecure download nasa nasa_meta.xml')
    assert files_downloaded('nasa') == {'nasa_meta.xml'}

    _stdout, stderr = call_cmd('ia --insecure download --checksum nasa nasa_meta.xml')
    assert files_downloaded('nasa') == {'nasa_meta.xml'}
    prefix = 'nasa:\n'.replace('\n', os.linesep)
    filepath = os.path.join('nasa', 'nasa_meta.xml')
    assert f'{prefix} skipping {filepath}, file already exists based on checksum.' == stderr


def test_checksum_archive(tmpdir_ch):
    call_cmd('ia --insecure download nasa nasa_meta.xml')
    assert files_downloaded('nasa') == {'nasa_meta.xml'}

    _stdout, stderr = call_cmd('ia --insecure download --checksum-archive nasa nasa_meta.xml')
    assert files_downloaded('nasa') == {'nasa_meta.xml'}
    prefix = 'nasa:\n'.replace('\n', os.linesep)
    filepath = os.path.join('nasa', 'nasa_meta.xml')
    assert f'{prefix} skipping {filepath}, file already exists based on checksum.' == stderr

    assert '_checksum_archive.txt' in files_downloaded('.')
    with open(os.path.join('.', '_checksum_archive.txt'), encoding='utf-8') as f:
        filepath = os.path.join('nasa', 'nasa_meta.xml')
        assert f.read() == f'{filepath}\n'

    _stdout, stderr = call_cmd('ia --insecure download --checksum-archive nasa nasa_meta.xml')
    assert files_downloaded('nasa') == {'nasa_meta.xml'}
    prefix = 'nasa:\n'.replace('\n', os.linesep)
    filepath = os.path.join('nasa', 'nasa_meta.xml')
    assert f'{prefix} skipping {filepath}, file already exists based on checksum_archive.' == stderr


def test_no_directories(tmpdir_ch):
    call_cmd('ia --insecure download --no-directories nasa nasa_meta.xml')
    assert files_downloaded('.') == {'nasa_meta.xml'}


def test_destdir(tmpdir_ch):
    cmd = 'ia --insecure download --destdir=thisdirdoesnotexist/ nasa nasa_meta.xml'
    _stdout, stderr = call_cmd(cmd, expected_exit_code=2)

    assert "--destdir: 'thisdirdoesnotexist/' is not a valid directory" in stderr

    tmpdir_ch.mkdir('thisdirdoesnotexist/')
    call_cmd(cmd)
    assert files_downloaded('thisdirdoesnotexist/nasa') == {'nasa_meta.xml'}

    tmpdir_ch.mkdir('dir2/')
    cmd = ('ia --insecure download --no-directories --destdir=dir2/ '
           'nasa nasa_meta.xml')
    call_cmd(cmd)
    assert files_downloaded('dir2') == {'nasa_meta.xml'}


def test_no_change_timestamp(tmpdir_ch):
    # TODO: Handle the case of daylight savings time

    now = time.time()
    call_cmd('ia --insecure download --no-change-timestamp nasa')

    for path, dirnames, filenames in os.walk(str(tmpdir_ch)):
        for d in dirnames:
            p = os.path.join(path, d)
            assert os.stat(p).st_mtime >= now

        for f in filenames:
            p = os.path.join(path, f)
            assert os.stat(p).st_mtime >= now


def test_download_history_flag(capsys):
    """Test that --download-history correctly includes/excludes history files.

    Regression test for https://github.com/jjjake/internetarchive/issues/735
    The bug was that --download-history was being passed directly to ignore_history_dir
    without negation, causing the opposite behavior.
    """
    # Add a history file to the nasa metadata
    nasa_data = json.loads(load_test_data_file('metadata/nasa.json'))
    nasa_data['files'].append({
        'name': 'history/files/old_file.txt',
        'source': 'original',
        'size': '100',
        'format': 'Text',
    })

    with IaRequestsMock() as mocker:
        mocker.add_metadata_mock('nasa', body=json.dumps(nasa_data))
        item = get_item('nasa')

        # Without --download-history (ignore_history_dir=True), history files excluded
        item.download(dry_run=True, ignore_history_dir=True)
        stdout_without = capsys.readouterr().out
        assert 'history/files/old_file.txt' not in stdout_without

        # With --download-history (ignore_history_dir=False), history files included
        item.download(dry_run=True, ignore_history_dir=False)
        stdout_with = capsys.readouterr().out
        assert 'history/files/old_file.txt' in stdout_with


def test_count_views_flag(tmpdir_ch):
    download_url_re = re.compile(r'https?://archive.org/download/.*')
    with IaRequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add_metadata_mock('nasa')
        rsps.add(responses.GET, download_url_re, body='test content')
        ia_call(['ia', '--insecure', 'download', '--no-directories',
                 '--count-views', 'nasa', 'nasa_meta.xml'])
        assert 'cnt=' not in rsps.calls[-1].request.url


def test_default_sends_cnt_zero(tmpdir_ch):
    download_url_re = re.compile(r'https?://archive.org/download/.*')
    with IaRequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add_metadata_mock('nasa')
        rsps.add(responses.GET, download_url_re, body='test content')
        ia_call(['ia', '--insecure', 'download', '--no-directories',
                 'nasa', 'nasa_meta.xml'])
        assert 'cnt=0' in rsps.calls[-1].request.url


@pytest.mark.parametrize(('value', 'expected'), [
    ('0-1023', 'bytes=0-1023'),
    ('bytes=0-1023', 'bytes=0-1023'),
    ('1024-', 'bytes=1024-'),
    ('-5', 'bytes=-5'),
    ('bytes=-1024', 'bytes=-1024'),
    (' 0-100 ', 'bytes=0-100'),
    ('BYTES=5-9', 'bytes=5-9'),
])
def test_normalize_byte_range_valid(value, expected):
    assert normalize_byte_range(value) == expected


@pytest.mark.parametrize('value', ['abc', '-', '5-1', '', 'bytes=', '0-1-2'])
def test_normalize_byte_range_invalid(value):
    with pytest.raises(ValueError, match='range'):
        normalize_byte_range(value)


@pytest.mark.parametrize(('value', 'expected'), [
    ('0-9', ['bytes=0-9']),
    ('0-9,50-99', ['bytes=0-9', 'bytes=50-99']),
    ('bytes=0-9,-100', ['bytes=0-9', 'bytes=-100']),
    (' 0-9 , 50-99 ', ['bytes=0-9', 'bytes=50-99']),
])
def test_parse_byte_ranges_valid(value, expected):
    assert parse_byte_ranges(value) == expected


@pytest.mark.parametrize('value', ['', '0-9,', ',', '0-9,abc', '0-9,,50-99'])
def test_parse_byte_ranges_invalid(value):
    with pytest.raises(ValueError, match='range'):
        parse_byte_ranges(value)


def _range_headers(rsps):
    """Return the Range header of every download request, in call order."""
    return [c.request.headers['Range']
            for c in rsps.calls if 'Range' in c.request.headers]


def test_range_single_bare(tmpdir_ch):
    download_url_re = re.compile(r'https?://archive.org/download/.*')
    with IaRequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add_metadata_mock('nasa')
        rsps.add(responses.GET, download_url_re, body='test')
        ia_call(['ia', '--insecure', 'download', '--no-directories', '--stdout',
                 '--range', '0-3', 'nasa', 'nasa_meta.xml'])
        assert _range_headers(rsps) == ['bytes=0-3']


def test_range_suffix_bare(tmpdir_ch):
    download_url_re = re.compile(r'https?://archive.org/download/.*')
    with IaRequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add_metadata_mock('nasa')
        rsps.add(responses.GET, download_url_re, body='test')
        ia_call(['ia', '--insecure', 'download', '--no-directories', '--stdout',
                 '--range', '-100', 'nasa', 'nasa_meta.xml'])
        assert _range_headers(rsps) == ['bytes=-100']


def test_range_bare_multi_range_one_file(tmpdir_ch):
    download_url_re = re.compile(r'https?://archive.org/download/.*')
    with IaRequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add_metadata_mock('nasa')
        rsps.add(responses.GET, download_url_re, body='AAAA')
        rsps.add(responses.GET, download_url_re, body='BBBB')
        ia_call(['ia', '--insecure', 'download', '--no-directories', '--stdout',
                 '--range', '0-3', '--range', '5-9', 'nasa', 'nasa_meta.xml'])
        # Both ranges, same file, in flag order.
        assert _range_headers(rsps) == ['bytes=0-3', 'bytes=5-9']


def test_range_bare_comma_one_file(tmpdir_ch):
    download_url_re = re.compile(r'https?://archive.org/download/.*')
    with IaRequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add_metadata_mock('nasa')
        rsps.add(responses.GET, download_url_re, body='AAAA')
        rsps.add(responses.GET, download_url_re, body='BBBB')
        ia_call(['ia', '--insecure', 'download', '--no-directories', '--stdout',
                 '--range', '0-3,5-9', 'nasa', 'nasa_meta.xml'])
        # A comma list expands to one request per range, in order -- identical
        # to repeating --range.
        assert _range_headers(rsps) == ['bytes=0-3', 'bytes=5-9']


def test_range_file_form_comma(tmpdir_ch):
    download_url_re = re.compile(r'https?://archive.org/download/.*')
    with IaRequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add_metadata_mock('nasa')
        rsps.add(responses.GET, download_url_re, body='AAAA')
        rsps.add(responses.GET, download_url_re, body='BBBB')
        ia_call(['ia', '--insecure', 'download', '--no-directories', '--stdout',
                 '--range', 'nasa_meta.xml:0-3,5-9', 'nasa'])
        assert _range_headers(rsps) == ['bytes=0-3', 'bytes=5-9']
        urls = [c.request.url for c in rsps.calls if 'Range' in c.request.headers]
        assert all('nasa_meta.xml' in u for u in urls)


def test_range_bare_single_range_multi_file(tmpdir_ch):
    download_url_re = re.compile(r'https?://archive.org/download/.*')
    with IaRequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add_metadata_mock('nasa')
        rsps.add(responses.GET, download_url_re, body='AAAA')
        rsps.add(responses.GET, download_url_re, body='BBBB')
        ia_call(['ia', '--insecure', 'download', '--no-directories', '--stdout',
                 '--range', '0-3', 'nasa', 'nasa_meta.xml', 'globe_west_540.jpg'])
        # Same range applied to each file, in file order.
        assert _range_headers(rsps) == ['bytes=0-3', 'bytes=0-3']
        urls = [c.request.url for c in rsps.calls if 'Range' in c.request.headers]
        assert 'nasa_meta.xml' in urls[0]
        assert 'globe_west_540.jpg' in urls[1]


def test_range_file_form_multi_file(tmpdir_ch):
    download_url_re = re.compile(r'https?://archive.org/download/.*')
    with IaRequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add_metadata_mock('nasa')
        rsps.add(responses.GET, download_url_re, body='AAAA')
        rsps.add(responses.GET, download_url_re, body='BBBB')
        ia_call(['ia', '--insecure', 'download', '--no-directories', '--stdout',
                 '--range', 'nasa_meta.xml:0-3',
                 '--range', 'globe_west_540.jpg:5-9', 'nasa'])
        assert _range_headers(rsps) == ['bytes=0-3', 'bytes=5-9']
        urls = [c.request.url for c in rsps.calls if 'Range' in c.request.headers]
        assert 'nasa_meta.xml' in urls[0]
        assert 'globe_west_540.jpg' in urls[1]


def test_range_file_form_same_file_repeated(tmpdir_ch):
    download_url_re = re.compile(r'https?://archive.org/download/.*')
    with IaRequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add_metadata_mock('nasa')
        rsps.add(responses.GET, download_url_re, body='AAAA')
        rsps.add(responses.GET, download_url_re, body='BBBB')
        ia_call(['ia', '--insecure', 'download', '--no-directories', '--stdout',
                 '--range', 'nasa_meta.xml:0-3',
                 '--range', 'nasa_meta.xml:5-9', 'nasa'])
        assert _range_headers(rsps) == ['bytes=0-3', 'bytes=5-9']


@pytest.mark.parametrize('argv', [
    # --range requires --stdout
    ['ia', '--insecure', 'download', 'nasa', 'nasa_meta.xml', '--range', '0-3'],
    # bare range with no file to bind to
    ['ia', '--insecure', 'download', '--stdout', 'nasa', '--range', '0-3'],
    # multiple bare ranges with multiple files is ambiguous
    ['ia', '--insecure', 'download', '--stdout', 'nasa', 'a', 'b',
     '--range', '0-3', '--range', '5-9'],
    # a comma list is multiple ranges too -- still ambiguous with multiple files
    ['ia', '--insecure', 'download', '--stdout', 'nasa', 'a', 'b',
     '--range', '0-3,5-9'],
    # cannot mix bare and FILE: forms
    ['ia', '--insecure', 'download', '--stdout', 'nasa',
     '--range', '0-3', '--range', 'f:5-9'],
    # FILE: form cannot combine with selectors
    ['ia', '--insecure', 'download', '--stdout', 'nasa',
     '--glob', '*', '--range', 'f:0-3'],
    ['ia', '--insecure', 'download', '--stdout', '--itemlist', '/dev/null',
     '--range', 'f:0-3'],
    # cannot read identifiers from stdin with --range
    ['ia', '--insecure', 'download', '--stdout', '-', '--range', 'f:0-3'],
    # invalid range value
    ['ia', '--insecure', 'download', '--stdout', 'nasa', 'nasa_meta.xml',
     '--range', 'not-a-range'],
])
def test_range_invalid_invocations(argv):
    ia_call(argv, expected_exit_code=2)


def test_range_job_failure_exits_nonzero(tmpdir_ch):
    """A failed range segment must propagate a nonzero exit code, so a downstream
    pipe consumer can tell the bytes are incomplete."""
    download_url_re = re.compile(r'https?://archive.org/download/.*')
    with IaRequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add_metadata_mock('nasa')
        rsps.add(responses.GET, download_url_re,
                 status=416,
                 adding_headers={'Content-Range': 'bytes */7105'})
        ia_call(['ia', '--insecure', 'download', '--no-directories', '--stdout',
                 '--range', '99999999-', 'nasa', 'nasa_meta.xml'],
                expected_exit_code=1)


def _range_args(ranges, identifier='nasa', file=None):
    """Build a minimal argparse.Namespace for build_range_jobs unit tests."""
    return argparse.Namespace(
        identifier=identifier, file=file or [], glob=None, format=None,
        source=None, exclude_source=None, search=None, itemlist=None,
        ranges=ranges,
    )


def test_build_range_jobs_filename_with_colon():
    """A FILE:RANGE value whose filename itself contains a colon binds to the
    full filename (split on the last colon)."""
    parser = argparse.ArgumentParser()
    args = _range_args(['weird:name.warc.gz:0-9'])
    assert build_range_jobs(args, parser) == [('weird:name.warc.gz', 'bytes=0-9')]



import argparse
import sys

from internetarchive.cli.ia_delete import get_files_to_delete


def make_args(**kwargs):
    args = dict(all=False, file=[], glob=None, format=None, cascade=False)  # noqa: C408
    args.update(kwargs)
    return argparse.Namespace(**args)


def test_get_files_to_delete_all(nasa_item):
    args = make_args(all=True)
    files = list(get_files_to_delete(args, nasa_item))
    expected = {
        'NASAarchiveLogo.jpg',
        'globe_west_540.jpg',
        'nasa_archive.torrent',
        'nasa_files.xml',
        'nasa_meta.xml',
        'nasa_reviews.xml',
    }
    assert {f.name for f in files} == expected


def test_get_files_to_delete_empty_file_list(nasa_item):
    args = make_args(file=[])
    files = list(get_files_to_delete(args, nasa_item))
    expected = {
        'NASAarchiveLogo.jpg',
        'globe_west_540.jpg',
        'nasa_archive.torrent',
        'nasa_files.xml',
        'nasa_meta.xml',
        'nasa_reviews.xml',
    }
    assert {f.name for f in files} == expected


def test_get_files_to_delete_with_glob(nasa_item):
    args = make_args(glob="*xml")
    files = list(get_files_to_delete(args, nasa_item))
    expected = {'nasa_meta.xml', 'nasa_reviews.xml', 'nasa_files.xml'}
    assert {f.name for f in files} == expected

    args = make_args(glob="*west_*")
    files = list(get_files_to_delete(args, nasa_item))
    expected = {'globe_west_540.jpg'}
    assert {f.name for f in files} == expected

    args = make_args(glob="*west_*|*torrent")
    files = list(get_files_to_delete(args, nasa_item))
    expected = {'globe_west_540.jpg', 'nasa_archive.torrent'}
    assert {f.name for f in files} == expected

    args = make_args(glob="nasa_[!m]*.xml")
    files = list(get_files_to_delete(args, nasa_item))
    expected = {'nasa_files.xml', 'nasa_reviews.xml'}
    assert {f.name for f in files} == expected

    args = make_args(glob="nasa_???????.xml")
    files = list(get_files_to_delete(args, nasa_item))
    expected = {'nasa_reviews.xml'}
    assert {f.name for f in files} == expected

    args = make_args(glob="*_[0-9]*")
    files = list(get_files_to_delete(args, nasa_item))
    expected = {'globe_west_540.jpg'}
    assert {f.name for f in files} == expected

    # Match JPEG files with uppercase letters in the name prefix
    args = make_args(glob="[A-Z]*.jpg")
    files = list(get_files_to_delete(args, nasa_item))
    expected = {'NASAarchiveLogo.jpg'}
    assert {f.name for f in files} == expected


    # Match lowercase-only names ending in .jpg
    args = make_args(glob="[a-z]*.jpg")
    files = list(get_files_to_delete(args, nasa_item))
    expected = {'globe_west_540.jpg'}
    assert {f.name for f in files} == expected

    args = make_args(glob="nasa_[fm]*.xml")
    files = list(get_files_to_delete(args, nasa_item))
    expected = {'nasa_files.xml', 'nasa_meta.xml'}
    assert {f.name for f in files} == expected


    args = make_args(glob="*.[a-z][a-z][a-z]")
    files = list(get_files_to_delete(args, nasa_item))
    expected = {
        'NASAarchiveLogo.jpg',
        'globe_west_540.jpg',
        'nasa_files.xml',
        'nasa_meta.xml',
        'nasa_reviews.xml'
    }
    assert {f.name for f in files} == expected

    args = make_args(glob="?a*")
    files = list(get_files_to_delete(args, nasa_item))
    expected = {
        'nasa_archive.torrent',
        'nasa_files.xml',
        'nasa_meta.xml',
        'nasa_reviews.xml'
    }
    assert {f.name for f in files} == expected

    args = make_args(glob="g*.jpg")
    files = list(get_files_to_delete(args, nasa_item))
    expected = {'globe_west_540.jpg'}
    assert {f.name for f in files} == expected

    args = make_args(glob="nasa_*[st].xml")
    files = list(get_files_to_delete(args, nasa_item))
    expected = {'nasa_files.xml', 'nasa_reviews.xml'}
    assert {f.name for f in files} == expected

    args = make_args(glob="[!nN]*")
    files = list(get_files_to_delete(args, nasa_item))
    expected = {'globe_west_540.jpg'}
    assert {f.name for f in files} == expected


def test_get_files_to_delete_with_format(nasa_item):
    args = make_args(format="JPEG")
    files = list(get_files_to_delete(args, nasa_item))
    expected = {'globe_west_540.jpg'}
    assert {f.name for f in files} == expected


def test_get_files_to_delete_with_explicit_file_list(nasa_item):
    args = make_args(file=["nasa_meta.xml", "nasa_reviews.xml"])
    files = list(get_files_to_delete(args, nasa_item))
    expected = {f.name for f in nasa_item.get_files(["nasa_meta.xml", "nasa_reviews.xml"])}
    assert {f.name for f in files} == expected


def test_get_files_to_delete_with_stdin(monkeypatch, nasa_item):
    args = make_args(file=["-"])
    monkeypatch.setattr(sys, "stdin", ["nasa_meta.xml\n", "nasa_reviews.xml\n"])
    files = list(get_files_to_delete(args, nasa_item))
    expected = {f.name for f in nasa_item.get_files(["nasa_meta.xml", "nasa_reviews.xml"])}
    assert {f.name for f in files} == expected



#
# The internetarchive module is a Python/CLI interface to Archive.org.
#
# Copyright (C) 2012-2026 Internet Archive
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
internetarchive.catalog
~~~~~~~~~~~~~~~~~~~~~~~

This module contains objects for interacting with the Archive.org catalog.

:copyright: (C) 2012-2024 by Internet Archive.
:license: AGPL 3, see LICENSE for more details.
"""
from __future__ import annotations

import time
from datetime import datetime
from logging import getLogger
from typing import Iterable, Iterator, Mapping, MutableMapping

import requests
from requests import Response
from requests.exceptions import HTTPError

from internetarchive import auth
from internetarchive import session as ia_session
from internetarchive.utils import json

log = getLogger(__name__)

FOLLOW_POLL_INTERVAL = 2.0

# How many consecutive transient request failures follow mode tolerates
# before re-raising (failures 1 through N-1 are retried; the Nth is fatal).
# Each counted failure already represents an exhausted session-layer urllib3
# retry cycle, so reaching this means a genuinely persistent outage.
FOLLOW_MAX_CONSECUTIVE_ERRORS = 5

# Task ``status`` values that mean the task is still alive and may produce more
# log output. A finished task is either done (returned from history with a null
# status) or errored (``status='error'``, which lingers in the catalog awaiting
# admin) -- both are terminal for follow purposes.
ACTIVE_TASK_STATUSES = frozenset({'running', 'queued', 'paused'})


def _is_transient_error(exc: requests.exceptions.RequestException) -> bool:
    """Return ``True`` if ``exc`` is a transient failure worth retrying.

    Connection errors, premature chunked responses, timeouts, and 5xx
    responses are transient. 4xx responses (and an ``HTTPError`` without an
    attached response) are fatal -- they won't heal by retrying.

    :param exc: The exception raised by a follow-mode request.

    :returns: ``True`` if the error is transient, else ``False``.
    """
    if isinstance(exc, HTTPError):
        return exc.response is not None and exc.response.status_code >= 500
    return isinstance(exc, (requests.exceptions.ConnectionError,
                            requests.exceptions.ChunkedEncodingError,
                            requests.exceptions.Timeout))


def sort_by_date(task_dict: CatalogTask) -> datetime:
    if task_dict.category == 'summary':  # type: ignore
        return datetime.now()
    try:
        return datetime.strptime(task_dict['submittime'], '%Y-%m-%d %H:%M:%S.%f')
    except Exception:
        return datetime.strptime(task_dict['submittime'], '%Y-%m-%d %H:%M:%S')


class Catalog:
    """This class represents the Archive.org catalog.
    You can use this class to access and submit tasks from the catalog.

    This is a low-level interface, and in most cases the functions
    in :mod:`internetarchive.api` and methods in
    :class:`ArchiveSession <ArchiveSession>` should be used.

    It uses the archive.org
    `Tasks API <https://archive.org/services/docs/api/tasks.html>`_

    Usage::
        >>> from internetarchive import get_session, Catalog
        >>> s = get_session()
        >>> c = Catalog(s)
        >>> tasks = c.get_tasks('nasa')
        >>> tasks[-1].task_id
        31643502
    """

    def __init__(
        self,
        archive_session: ia_session.ArchiveSession,
        request_kwargs: Mapping | None = None,
    ):
        """
        Initialize :class:`Catalog <Catalog>` object.

        :param archive_session: An :class:`ArchiveSession <ArchiveSession>`
                                object.

        :param request_kwargs: Keyword arguments to be used
                               in :meth:`requests.sessions.Session.get`
                               and :meth:`requests.sessions.Session.post`
                               requests.
        """
        self.session = archive_session
        self.auth = auth.S3Auth(self.session.access_key, self.session.secret_key)
        self.request_kwargs = request_kwargs or {}
        self.url = f'{self.session.protocol}//{self.session.host}/services/tasks.php'

    def get_summary(self, identifier: str = "", params: dict | None = None) -> dict:
        """Get the total counts of catalog tasks meeting all criteria,
        organized by run status (queued, running, error, and paused).


        :param identifier: Item identifier.

        :param params: Query parameters, refer to

        `Tasks API <https://archive.org/services/docs/api/tasks.html>`_
        for available parameters.

        :returns: the total counts of catalog tasks meeting all criteria
        """
        params = params or {}
        if identifier:
            params['identifier'] = identifier
        params.update({'summary': 1, 'history': 0, 'catalog': 0})
        r = self.make_tasks_request(params)
        j = r.json()
        if j.get('success') is True:
            return j['value']['summary']
        else:
            return j

    def make_tasks_request(self, params: Mapping | None) -> Response:
        """Make a GET request to the
         `Tasks API <https://archive.org/services/docs/api/tasks.html>`_

        :param params: Query parameters, refer to
                       `Tasks API
                       <https://archive.org/services/docs/api/tasks.html>`_
                       for available parameters.

        :returns: :class:`requests.Response`
        """
        r = self.session.get(self.url,
                             params=params,
                             auth=self.auth,
                             **self.request_kwargs)
        try:
            r.raise_for_status()
        except HTTPError as exc:
            j = r.json()
            error = j['error']
            raise HTTPError(error, response=r)
        return r

    def iter_tasks(self, params: MutableMapping | None = None) -> Iterable[CatalogTask]:
        """A generator that can make arbitrary requests to the
        Tasks API. It handles paging (via cursor) automatically.

        :param params: Query parameters, refer to
                       `Tasks API
                       <https://archive.org/services/docs/api/tasks.html>`_
                       for available parameters.

        :returns: collections.Iterable[CatalogTask]
        """
        params = params or {}
        while True:
            r = self.make_tasks_request(params)
            j = r.json()
            for row in j.get('value', {}).get('catalog', []):
                yield CatalogTask(row, self)
            for row in j.get('value', {}).get('history', []):
                yield CatalogTask(row, self)
            if not j.get('value', {}).get('cursor'):
                break
            params['cursor'] = j['value']['cursor']

    def get_rate_limit(self, cmd: str = 'derive.php') -> dict:
        """Get the current rate limit status for a task command.

        :param cmd: The task command to check (e.g., ``'derive.php'``).

        :returns: A dict containing rate limit information.
        """
        params = {'rate_limits': 1, 'cmd': cmd}
        r = self.make_tasks_request(params)
        line = ''
        tasks = []
        for c in r.iter_content():
            c = c.decode('utf-8')
            if c == '\n':
                j = json.loads(line)
                task = CatalogTask(j, self)
                tasks.append(task)
                line = ''
            line += c
        j = json.loads(line)
        return j

    def get_tasks(self, identifier: str = "", params: dict | None = None) -> list[CatalogTask]:
        """Get a list of all tasks meeting all criteria.
        The list is ordered by submission time.

        :param identifier: The item identifier, if provided
                           will return tasks for only this item filtered by
                           other criteria provided in params.

        :param params: Query parameters, refer to

        `Tasks API <https://archive.org/services/docs/api/tasks.html>`_
        for available parameters.

        :returns: A list of all tasks meeting all criteria.
        """
        params = params or {}
        if identifier:
            params.update({'identifier': identifier})
        params.update({'limit': 0})
        if not params.get('summary'):
            params['summary'] = 0
        r = self.make_tasks_request(params)
        line = ''
        tasks = []
        for c in r.iter_content():
            c = c.decode('utf-8')
            if c == '\n':
                j = json.loads(line)
                task = CatalogTask(j, self)
                tasks.append(task)
                line = ''
            line += c
        if line.strip():
            j = json.loads(line)
            task = CatalogTask(j, self)
            tasks.append(task)

        all_tasks = sorted(tasks, key=sort_by_date, reverse=True)
        return all_tasks

    def submit_task(self, identifier: str, cmd: str,
                    comment: str | None = None,
                    priority: int = 0,
                    data: dict | None = None,
                    headers: dict | None = None) -> Response:
        """Submit an archive.org task.

        :param identifier: Item identifier.

        :param cmd: Task command to submit, see
                    `supported task commands
                    <https://archive.org/services/docs/api/tasks.html#supported-tasks>`_.

        :param comment: A reasonable explanation for why the
                        task is being submitted.

        :param priority: Task priority from 10 to -10
                         (default: 0).

        :param data: Extra POST data to submit with
                     the request. Refer to `Tasks API Request Entity
                     <https://archive.org/services/docs/api/tasks.html#request-entity>`_.

        :param headers: Add additional headers to request.

        :returns: :class:`requests.Response`
        """
        data = data or {}
        data.update({'cmd': cmd, 'identifier': identifier})
        if comment:
            if 'args' in data:
                data['args']['comment'] = comment
            else:
                data['args'] = {'comment': comment}
        if priority:
            data['priority'] = priority
        r = self.session.post(self.url,
                              json=data,
                              auth=self.auth,
                              headers=headers,
                              **self.request_kwargs)
        return r


class CatalogTask:
    """This class represents an Archive.org catalog task. It is primarily used by
    :class:`Catalog`, and should not be used directly.
    """
    def __init__(self, task_dict: Mapping, catalog_obj: Catalog):
        self.session = catalog_obj.session
        self.request_kwargs = catalog_obj.request_kwargs
        self.color = None
        self.task_dict = task_dict
        for key, value in task_dict.items():
            setattr(self, key, value)  # Confuses mypy ;-)

    def __repr__(self):
        color = self.task_dict.get('color', 'done')
        return ('CatalogTask(identifier={identifier},'
                ' task_id={task_id!r}, server={server!r},'
                ' cmd={cmd!r},'
                ' submitter={submitter!r},'
                ' color={task_color!r})'.format(task_color=color, **self.task_dict))

    def __getitem__(self, key: str):
        """Dict-like access provided as backward compatibility."""
        return self.task_dict[key]

    def json(self):
        return json.dumps(self.task_dict)

    def task_log(self) -> str:
        """Get task log.

        :returns: The task log as a string.

        """
        task_id = self.task_id  # type: ignore
        if task_id is None:
            raise ValueError('task_id is None')
        return self.get_task_log(task_id, self.session,
                                 request_kwargs=self.request_kwargs)

    @staticmethod
    def _request_task_log(
        task_id: int | str | None,
        session: ia_session.ArchiveSession,
        *,
        params: Mapping | None = None,
        request_kwargs: Mapping | None = None
    ) -> Response:
        """Make the HTTP request for a task log and return the raw Response.

        :param task_id: The task id for the task log you'd like to fetch.

        :param session: :class:`ArchiveSession <ArchiveSession>`

        :param params: Extra URL parameters to send with the request.

        :param request_kwargs: Keyword arguments that
                               :py:class:`requests.Request` takes.

        :returns: :class:`requests.Response`
        """
        request_kwargs = request_kwargs or {}
        _auth = auth.S3Auth(session.access_key, session.secret_key)
        if session.host == 'archive.org':
            host = 'catalogd.archive.org'
        else:
            host = session.host
        url = f'{session.protocol}//{host}/services/tasks.php'
        _params = {'task_log': task_id}
        if params:
            _params.update(params)
        r = session.get(url, params=_params, auth=_auth, **request_kwargs)
        r.raise_for_status()
        return r

    @staticmethod
    def get_task_log(
        task_id: int | str | None,
        session: ia_session.ArchiveSession,
        *,
        params: Mapping | None = None,
        request_kwargs: Mapping | None = None
    ) -> str:
        """Static method for getting a task log, given a task_id.

        This method exists so a task log can be retrieved without
        retrieving the items task history first.

        :param task_id: The task id for the task log you'd like to fetch.

        :param session: :class:`ArchiveSession <ArchiveSession>`

        :param params: URL parameters to send with the request (e.g.
                       ``{'lines': 100}`` to fetch a truncated log).

        :param request_kwargs: Keyword arguments that
                               :py:class:`requests.Request` takes.

        :returns: The task log as a string.
        """
        r = CatalogTask._request_task_log(task_id, session, params=params,
                                          request_kwargs=request_kwargs)
        return r.content.decode('utf-8', errors='surrogateescape')

    @staticmethod
    def _select_log_lines(text: str, lines: int | None) -> str:
        """Return the initial backlog slice of ``text`` for follow mode.

        Replicates Tasks API ``lines`` semantics: ``None`` -> whole log,
        negative ``N`` -> last N lines, ``0`` -> nothing. Positive values
        (first N lines / the head of the log) are rejected by the callers,
        since a head slice cannot be followed.

        :param text: The full decoded task log.

        :param lines: Number of lines of backlog to keep, or ``None`` for all.

        :returns: The selected backlog text (newlines preserved).
        """
        if lines is None:
            return text
        if lines == 0:
            return ''
        parts = text.splitlines(keepends=True)
        selected = parts[:lines] if lines > 0 else parts[lines:]
        return ''.join(selected)

    @staticmethod
    def _task_is_active(
        task_id: int | str,
        session: ia_session.ArchiveSession,
        request_kwargs: Mapping | None = None
    ) -> bool:
        """Return ``True`` while the task may still produce log output.

        A task's ``status`` is the authoritative signal: ``running``,
        ``queued`` and ``paused`` are alive, while ``error`` (which lingers in
        the catalog awaiting admin) and ``done`` (returned from history with a
        null status) are terminal. Catalog membership alone is not reliable.

        :param task_id: The task id to check.

        :param session: :class:`ArchiveSession <ArchiveSession>`

        :param request_kwargs: Keyword arguments for the request.

        :returns: ``True`` if the task is active, else ``False``.
        """
        tasks = session.get_tasks(
            params={'task_id': task_id, 'catalog': 1, 'history': 1, 'summary': 0},
            request_kwargs=request_kwargs)
        for t in tasks:
            if str(getattr(t, 'task_id', '')) == str(task_id):
                if getattr(t, 'status', None) in ACTIVE_TASK_STATUSES:
                    return True
        return False

    @staticmethod
    def follow_task_log(
        task_id: int | str,
        session: ia_session.ArchiveSession,
        *,
        lines: int | None = None,
        params: Mapping | None = None,
        request_kwargs: Mapping | None = None
    ) -> Iterator[str]:
        """Follow a task log as it grows, ``tail -f`` style.

        Yields newly appended text as it appears. Stops automatically when
        the task's status indicates it has finished (the task is no longer
        ``running``, ``queued``, or ``paused``).

        :param task_id: The task id to follow.

        :param session: :class:`ArchiveSession <ArchiveSession>`

        :param lines: How much existing backlog to emit before following,
                      using Tasks API ``lines`` semantics: ``None`` = the
                      whole log, negative ``N`` = the last ``N`` lines, ``0`` =
                      none. A positive value (the head of the log) cannot be
                      followed and is rejected.

        :param params: Extra URL parameters forwarded to every task-log
                       request. Do not pass the Tasks API ``lines`` parameter
                       here (it would truncate the body and break delta
                       tracking) -- use the ``lines`` argument instead.

        :param request_kwargs: Keyword arguments that
                               :py:class:`requests.Request` takes.

        :returns: An iterator of newly appended log text.

        :raises ValueError: If ``lines`` is positive (the head of the log
            cannot be followed; use a negative value for the last N lines).

        :raises requests.exceptions.RequestException: On a fatal request
            error, or when transient errors persist for
            ``FOLLOW_MAX_CONSECUTIVE_ERRORS`` consecutive polls.
        """
        if lines is not None and lines > 0:
            raise ValueError(
                "follow_task_log: a positive 'lines' value selects the head "
                "of the log and cannot be followed; use a negative value for "
                "the last N lines (e.g. lines=-20)")
        seen = 0
        last_modified = None
        first = True
        consecutive_errors = 0

        def _request_kwargs(conditional: bool) -> dict:
            # The conditional poll sends If-Modified-Since for a cheap 304;
            # the final/unconditional read must never be short-circuited by a
            # conditional header (including one a caller supplied), or it could
            # 304 away the trailing bytes it exists to catch.
            rk = dict(request_kwargs or {})
            headers = dict(rk.get('headers', {}))
            if conditional and last_modified:
                headers['If-Modified-Since'] = last_modified
            else:
                headers.pop('If-Modified-Since', None)
            rk['headers'] = headers
            return rk

        while True:
            # Everything is computed inside the try and yielded outside it,
            # so a failure can never lose or duplicate already-fetched
            # output, and exceptions never fire mid-yield.
            to_yield: list[str] = []
            stop = False
            try:
                checked_first = first
                r = CatalogTask._request_task_log(
                    task_id, session, params=params,
                    request_kwargs=_request_kwargs(conditional=True))

                grew = False
                if r.status_code != 304:
                    body = r.content.decode('utf-8', errors='surrogateescape')
                    lm = r.headers.get('Last-Modified')
                    if lm:
                        last_modified = lm
                    # Task logs are append-only in practice; a body shorter
                    # than what we've already emitted can only be a transient
                    # bad response, so such a poll is skipped entirely.
                    if first:
                        initial = CatalogTask._select_log_lines(body, lines)
                        if initial:
                            to_yield.append(initial)
                        seen = len(body)
                        first = False
                        grew = True
                    elif len(body) >= seen:
                        new = body[seen:]
                        if new:
                            to_yield.append(new)
                            seen = len(body)
                            grew = True

                # Check the task status after the first poll (so an already
                # finished task exits at once instead of waiting a full poll
                # interval) and on any later poll that produced no new output.
                terminal = False
                if checked_first or not grew:
                    terminal = not CatalogTask._task_is_active(
                        task_id, session, request_kwargs=request_kwargs)
                # Only fetch the trailing bytes when this poll produced nothing,
                # so a failed final fetch can never discard buffered output.
                if terminal and not grew:
                    r = CatalogTask._request_task_log(
                        task_id, session, params=params,
                        request_kwargs=_request_kwargs(conditional=False))
                    body = r.content.decode('utf-8', errors='surrogateescape')
                    if len(body) >= seen:
                        new = body[seen:]
                        if new:
                            to_yield.append(new)
                    stop = True
            except requests.exceptions.RequestException as exc:
                if not _is_transient_error(exc):
                    raise
                consecutive_errors += 1
                if consecutive_errors >= FOLLOW_MAX_CONSECUTIVE_ERRORS:
                    raise
                log.warning('Transient error following task %s log (%d/%d): %s',
                            task_id, consecutive_errors,
                            FOLLOW_MAX_CONSECUTIVE_ERRORS, exc)
                time.sleep(FOLLOW_POLL_INTERVAL)
                continue
            consecutive_errors = 0
            yield from to_yield
            if stop:
                return
            # A terminal task that just emitted its backlog reconciles on the
            # next poll without sleeping first -- no idle wait before exit.
            if not terminal:
                time.sleep(FOLLOW_POLL_INTERVAL)
